from typing import List
import grpc
from ..dispatcher_pb2_grpc import DispatcherStub

from ..dispatcher_pb2 import SubscribeToWorkflowEventsRequest, ResourceEventType
from ..loader import ClientConfig
from ..metadata import get_metadata
import json
import time


DEFAULT_ACTION_LISTENER_RETRY_INTERVAL = 5  # seconds
DEFAULT_ACTION_LISTENER_RETRY_COUNT = 5


class StepRunEventType:
    STEP_RUN_EVENT_TYPE_STARTED = 'STEP_RUN_EVENT_TYPE_STARTED'
    STEP_RUN_EVENT_TYPE_COMPLETED = 'STEP_RUN_EVENT_TYPE_COMPLETED'
    STEP_RUN_EVENT_TYPE_FAILED = 'STEP_RUN_EVENT_TYPE_FAILED'
    STEP_RUN_EVENT_TYPE_CANCELLED = 'STEP_RUN_EVENT_TYPE_CANCELLED'
    STEP_RUN_EVENT_TYPE_TIMED_OUT = 'STEP_RUN_EVENT_TYPE_TIMED_OUT'


event_type_mapping = {
    ResourceEventType.RESOURCE_EVENT_TYPE_STARTED: StepRunEventType.STEP_RUN_EVENT_TYPE_STARTED,
    ResourceEventType.RESOURCE_EVENT_TYPE_COMPLETED: StepRunEventType.STEP_RUN_EVENT_TYPE_COMPLETED,
    ResourceEventType.RESOURCE_EVENT_TYPE_FAILED: StepRunEventType.STEP_RUN_EVENT_TYPE_FAILED,
    ResourceEventType.RESOURCE_EVENT_TYPE_CANCELLED: StepRunEventType.STEP_RUN_EVENT_TYPE_CANCELLED,
    ResourceEventType.RESOURCE_EVENT_TYPE_TIMED_OUT: StepRunEventType.STEP_RUN_EVENT_TYPE_TIMED_OUT,
}


class StepRunEvent:
    def __init__(self, type: StepRunEventType, payload: str):
        self.type = type
        self.payload = payload


def new_listener(conn, config: ClientConfig):
    return ListenerClientImpl(
        client=DispatcherStub(conn),
        token=config.token,
    )


class ListenerClientImpl:
    def __init__(self, client: DispatcherStub, token):
        self.client = client
        self.token = token

    async def generator(self, workflowRunId: str) -> List[StepRunEvent]:
        listener = self.retry_subscribe(workflowRunId)

        while True:
            try:
                for workflow_event in listener:
                    eventType = None

                    if workflow_event.eventType in event_type_mapping:
                        eventType = event_type_mapping[workflow_event.eventType]
                    else:
                        raise Exception(
                            f"Unknown event type: {workflow_event.eventType}")

                    payload = None
                    if workflow_event.eventPayload:
                        payload = json.loads(workflow_event.eventPayload)

                    # call the handler
                    event = StepRunEvent(type=eventType, payload=payload)
                    yield event

            except grpc.RpcError as e:
                # Handle different types of errors
                if e.code() == grpc.StatusCode.CANCELLED:
                    # Context cancelled, unsubscribe and close
                    break
                elif e.code() == grpc.StatusCode.UNAVAILABLE:
                    # Retry logic
                    # logger.info("Could not connect to Hatchet, retrying...")
                    listener = self.retry_subscribe(workflowRunId)
                elif e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                    # logger.info("Deadline exceeded, retrying subscription")
                    continue
                else:
                    # Unknown error, report and break
                    # logger.error(f"Failed to receive message: {e}")
                    break

    def on(self, workflowRunId: str, handler: callable = None):
        for event in self.generator(workflowRunId):
            # call the handler if provided
            if handler:
                handler(event)

    def retry_subscribe(self, workflowRunId: str):
        retries = 0

        while retries < DEFAULT_ACTION_LISTENER_RETRY_COUNT:
            try:
                if retries > 0:
                    time.sleep(DEFAULT_ACTION_LISTENER_RETRY_INTERVAL)

                listener = self.client.SubscribeToWorkflowEvents(
                    SubscribeToWorkflowEventsRequest(
                        workflowRunId=workflowRunId,
                    ), metadata=get_metadata(self.token))
                return listener
            except grpc.RpcError as e:
                if e.code() == grpc.StatusCode.UNAVAILABLE:
                    retries = retries + 1
                else:
                    raise ValueError(f"gRPC error: {e}")



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import dispatcher_pb2 as dispatcher__pb2


class DispatcherStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Register = channel.unary_unary(
            '/Dispatcher/Register',
            request_serializer=dispatcher__pb2.WorkerRegisterRequest.SerializeToString,
            response_deserializer=dispatcher__pb2.WorkerRegisterResponse.FromString,
        )
        self.Listen = channel.unary_stream(
            '/Dispatcher/Listen',
            request_serializer=dispatcher__pb2.WorkerListenRequest.SerializeToString,
            response_deserializer=dispatcher__pb2.AssignedAction.FromString,
        )
        self.SubscribeToWorkflowEvents = channel.unary_stream(
            '/Dispatcher/SubscribeToWorkflowEvents',
            request_serializer=dispatcher__pb2.SubscribeToWorkflowEventsRequest.SerializeToString,
            response_deserializer=dispatcher__pb2.WorkflowEvent.FromString,
        )
        self.SendStepActionEvent = channel.unary_unary(
            '/Dispatcher/SendStepActionEvent',
            request_serializer=dispatcher__pb2.StepActionEvent.SerializeToString,
            response_deserializer=dispatcher__pb2.ActionEventResponse.FromString,
        )
        self.SendGroupKeyActionEvent = channel.unary_unary(
            '/Dispatcher/SendGroupKeyActionEvent',
            request_serializer=dispatcher__pb2.GroupKeyActionEvent.SerializeToString,
            response_deserializer=dispatcher__pb2.ActionEventResponse.FromString,
        )
        self.Unsubscribe = channel.unary_unary(
            '/Dispatcher/Unsubscribe',
            request_serializer=dispatcher__pb2.WorkerUnsubscribeRequest.SerializeToString,
            response_deserializer=dispatcher__pb2.WorkerUnsubscribeResponse.FromString,
        )


class DispatcherServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Register(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Listen(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubscribeToWorkflowEvents(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendStepActionEvent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendGroupKeyActionEvent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Unsubscribe(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DispatcherServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'Register': grpc.unary_unary_rpc_method_handler(
            servicer.Register,
            request_deserializer=dispatcher__pb2.WorkerRegisterRequest.FromString,
            response_serializer=dispatcher__pb2.WorkerRegisterResponse.SerializeToString,
        ),
        'Listen': grpc.unary_stream_rpc_method_handler(
            servicer.Listen,
            request_deserializer=dispatcher__pb2.WorkerListenRequest.FromString,
            response_serializer=dispatcher__pb2.AssignedAction.SerializeToString,
        ),
        'SubscribeToWorkflowEvents': grpc.unary_stream_rpc_method_handler(
            servicer.SubscribeToWorkflowEvents,
            request_deserializer=dispatcher__pb2.SubscribeToWorkflowEventsRequest.FromString,
            response_serializer=dispatcher__pb2.WorkflowEvent.SerializeToString,
        ),
        'SendStepActionEvent': grpc.unary_unary_rpc_method_handler(
            servicer.SendStepActionEvent,
            request_deserializer=dispatcher__pb2.StepActionEvent.FromString,
            response_serializer=dispatcher__pb2.ActionEventResponse.SerializeToString,
        ),
        'SendGroupKeyActionEvent': grpc.unary_unary_rpc_method_handler(
            servicer.SendGroupKeyActionEvent,
            request_deserializer=dispatcher__pb2.GroupKeyActionEvent.FromString,
            response_serializer=dispatcher__pb2.ActionEventResponse.SerializeToString,
        ),
        'Unsubscribe': grpc.unary_unary_rpc_method_handler(
            servicer.Unsubscribe,
            request_deserializer=dispatcher__pb2.WorkerUnsubscribeRequest.FromString,
            response_serializer=dispatcher__pb2.WorkerUnsubscribeResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'Dispatcher', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))

 # This class is part of an EXPERIMENTAL API.


class Dispatcher(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Register(request,
                 target,
                 options=(),
                 channel_credentials=None,
                 call_credentials=None,
                 insecure=False,
                 compression=None,
                 wait_for_ready=None,
                 timeout=None,
                 metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Dispatcher/Register',
                                             dispatcher__pb2.WorkerRegisterRequest.SerializeToString,
                                             dispatcher__pb2.WorkerRegisterResponse.FromString,
                                             options, channel_credentials,
                                             insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Listen(request,
               target,
               options=(),
               channel_credentials=None,
               call_credentials=None,
               insecure=False,
               compression=None,
               wait_for_ready=None,
               timeout=None,
               metadata=None):
        return grpc.experimental.unary_stream(request, target, '/Dispatcher/Listen',
                                              dispatcher__pb2.WorkerListenRequest.SerializeToString,
                                              dispatcher__pb2.AssignedAction.FromString,
                                              options, channel_credentials,
                                              insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SubscribeToWorkflowEvents(request,
                                  target,
                                  options=(),
                                  channel_credentials=None,
                                  call_credentials=None,
                                  insecure=False,
                                  compression=None,
                                  wait_for_ready=None,
                                  timeout=None,
                                  metadata=None):
        return grpc.experimental.unary_stream(request, target, '/Dispatcher/SubscribeToWorkflowEvents',
                                              dispatcher__pb2.SubscribeToWorkflowEventsRequest.SerializeToString,
                                              dispatcher__pb2.WorkflowEvent.FromString,
                                              options, channel_credentials,
                                              insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendStepActionEvent(request,
                            target,
                            options=(),
                            channel_credentials=None,
                            call_credentials=None,
                            insecure=False,
                            compression=None,
                            wait_for_ready=None,
                            timeout=None,
                            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Dispatcher/SendStepActionEvent',
                                             dispatcher__pb2.StepActionEvent.SerializeToString,
                                             dispatcher__pb2.ActionEventResponse.FromString,
                                             options, channel_credentials,
                                             insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendGroupKeyActionEvent(request,
                                target,
                                options=(),
                                channel_credentials=None,
                                call_credentials=None,
                                insecure=False,
                                compression=None,
                                wait_for_ready=None,
                                timeout=None,
                                metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Dispatcher/SendGroupKeyActionEvent',
                                             dispatcher__pb2.GroupKeyActionEvent.SerializeToString,
                                             dispatcher__pb2.ActionEventResponse.FromString,
                                             options, channel_credentials,
                                             insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Unsubscribe(request,
                    target,
                    options=(),
                    channel_credentials=None,
                    call_credentials=None,
                    insecure=False,
                    compression=None,
                    wait_for_ready=None,
                    timeout=None,
                    metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Dispatcher/Unsubscribe',
                                             dispatcher__pb2.WorkerUnsubscribeRequest.SerializeToString,
                                             dispatcher__pb2.WorkerUnsubscribeResponse.FromString,
                                             options, channel_credentials,
                                             insecure, call_credentials, compression, wait_for_ready, timeout, metadata)
